/* bdb-LU.cp.c */

/* This program is the host processor software for a parallel version
of the Shur complement LU factorization algorithm for
block-diagonal-bordered form sparse matrices.  */

/* special considerations for this implementation: 

1) Matrices are in symmetric block-diagonal-bordered form

2) any number of processors can be specified up to an including the
total number of processors on the CM-5 

3) it is assumed that ordering and load balancing are performed a
priori

4) the last block is considered sparse and solved with a sparse
version of kij-saxpy LU factorization


Naming convention used throughout this software

i - row number
j - column number
k - number along the diagonal

Lij   - any value in a strictly lower diagonal matrix block
Lkk   - value of the kth element on the lower diagonal
        stored in the location diag[k].Value
Lix   - row in a strictly lower diagonal matrix block
Lxk   - column in a strictly lower diagonal matrix block
        only accessed when updating a column
LBix  - partial row in a lower matrix border 
LBxk  - column in a lower matrix border
        only accessed when updating a column

Uij   - any value in a strictly upper diagonal matrix block
Ukk   - value of the kth element on the upper diagonal
        always equal to 1.0, so it is not expliticly stored
Uxj   - column in a strictly upper diagonal matrix block
Ukx   - row in a strictly upper diagonal matrix block
        only accessed when updating a row
UBxj  - partial column in an upper matrix border
UBkx  - row in an upper matrix border
        only accessed when updating a row

Uij   - any value in a strictly upper diagonal matrix block
Ukk   - value of the kth element on the upper diagonal
        always equal to 1.0, so it is not expliticly stored
Uxj   - column in a strictly upper diagonal matrix block
Ukx   - row in a strictly upper diagonal matrix block
        only accessed when updating a row
UBxj  - partial column in an upper matrix border
UBkx  - row in an upper matrix border
        only accessed when updating a row
*/
/*---------------------------------------------------------*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <malloc.h>
#include <math.h>
#include <cm/cmmd.h>
#include "bdb_LU_struct.h"
#include "bdb_LU_MSG_NUM.h"
/*---------------------------------------------------------*/
#define YES        1
#define NO         0
/*---------------------------------------------------------*/
typedef struct Node_Timer{int pn_id;
                          double elapsed;
                          double busy;
                          double idle} node_timer;
node_timer node_tim;

typedef struct CMMD_Timer{double min;
                          double ave;
                          double max;
                          double last;} CMMD_Timer;
CMMD_Timer *tSTAT;  /* tSTAT[] */
CMMD_Timer tFAC,tFR,tBS,tTOT;
/*---------------------------------------------------------*/
BBuf bbuf;
int          n;          /* dimension of block-diagonal block */
int          np;         /* number of diagonal elements distributed to a processor */
int          nlb;        /* dimension of the last block */
int          nb;         /* block size */
/*---------------------------------------------------------*/
Proc        *proc;       /* proc[]      */
int          nprocs;
/*---------------------------------------------------------*/
Block       *blocks;     /* blocks[]     */
int          nblocks;    
Block       *pblocks;    /* pblocks[]    */
int          npblocks;    
/*---------------------------------------------------------*/
Diag        *diag;       /* diag[]       */
int          ndiag;
Diag        *pdiag;      /* pdiag[]      */
int          npdiag;
/*---------------------------------------------------------*/
BlockRow    *blockrow;   /* blockrow[]   */
int          nblockrow;
BlockRow    *pblockrow;  /* pblockrow[]  */
int          npblockrow;
/*---------------------------------------------------------*/
BlockCol    *blockcol;   /* blockcol[]   */
int          nblockcol;
BlockCol    *pblockcol;  /* pblockcol[]  */
int          npblockcol;
/*---------------------------------------------------------*/
LowerBorder *lborder;    /* lborder[]    */
int          nlborder;
LowerBorder *plborder;   /* plborder[]   */
int          nplborder;
/*---------------------------------------------------------*/
UpperBorder *uborder;    /* uborder[]    */
int          nuborder;
UpperBorder *puborder;   /* puborder[]   */
int          npuborder;
/*---------------------------------------------------------*/
BorderRow   *borderrow;  /* borderrow[]  */
int          nborderrow;
BorderRow   *pborderrow; /* pborderrow[] */
int          npborderrow;
/*---------------------------------------------------------*/
BorderCol   *bordercol;  /* bordercol[]  */
int          nbordercol;
BorderCol   *pbordercol; /* pbordercol[] */
int          npbordercol;
/*---------------------------------------------------------*/
int         *lbbufsizeFac;
int          nlbbufsizeFac;
int         *lbbufsizeFr;
int          nlbbufsizeFr;
/*---------------------------------------------------------*/
int          nlbdiag;
LBDiag      *plbdiag;
int          nplbdiag;
/*---------------------------------------------------------*/
LBRow       *plbrowU;
int          nplbrowU;
LBRow       *plbrowL;
int          nplbrowL;
/*---------------------------------------------------------*/
LBCol       *plbcolU;
int          nplbcolU;
LBCol       *plbcolL;
int          nplbcolL;
/*---------------------------------------------------------*/
LBDiagColPt  *lbdiagcolpt;
int          nlbdiagcolpt;
LBDiagColPt  *plbdiagcolpt;
int          nplbdiagcolpt;
/*---------------------------------------------------------*/
double         *b;         /* b[]         */
double         *pb;        /* pb[]        */
double         *lb;        /* lb[]        */
double         *plb;       /* plb[]       */
/*---------------------------------------------------------*/
int            bytesreceived;
int            bytessent;
/*---------------------------------------------------------*/
int    *nupdatesULB;
int     nnupdatesULB;
int    *nupdatesFRB;
int     nnupdatesFRB;
/*---------------------------------------------------------*/
/* function prototypes */
void   InputData();
int    min();

main(argc,argv)
  int argc;
  char **argv;
{
  int i,j,l,bufsize,nodesperproc;

  printf("\n Executing host program... (pid=%d)\n", getpid());

  if (argc < 2)
   {
    printf("Usage: (1) input_file_name_A_matrix \n");
    printf("       (2) input_file_name_b_vector \n");
    exit(0);
   } /* end if */

  CMMD_enable();

/* input the data from the file - allocate memory as required */
  printf("input file for the A matrix  = |%s|\n",argv[1]);
  printf("input file for b vector  = |%s|\n",argv[2]);
  fflush(stdout);

printf("InputData() \n");
fflush(stdout);
  InputData(argv[1],argv[2]);

printf("FactorIndependentBlocks() \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_FIB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[0].min) tSTAT[0].min = node_tim.busy;
    if (node_tim.busy > tSTAT[0].max) tSTAT[0].max = node_tim.busy;
    tSTAT[0].ave += node_tim.busy;
   } /* end for i */
  tSTAT[0].last = node_tim.busy;
  tSTAT[0].ave /= (double) nprocs;
  tFAC.min  += tSTAT[0].min;
  tFAC.max  += tSTAT[0].max;
  tFAC.ave  += tSTAT[0].ave;
  tFAC.last += tSTAT[0].last;
  tTOT.min  += tSTAT[0].min;
  tTOT.max  += tSTAT[0].max;
  tTOT.ave  += tSTAT[0].ave;
  tTOT.last += tSTAT[0].last;

printf("UpdateLastBlock() \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_ULB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[1].min) tSTAT[1].min = node_tim.busy;
    if (node_tim.busy > tSTAT[1].max) tSTAT[1].max = node_tim.busy;
    tSTAT[1].ave += node_tim.busy;
  } /* end for i */
  tSTAT[1].last = node_tim.busy;
  tSTAT[1].ave /= (double) nprocs;
  tFAC.min  += tSTAT[1].min;
  tFAC.max  += tSTAT[1].max;
  tFAC.ave  += tSTAT[1].ave;
  tFAC.last += tSTAT[1].last;
  tTOT.min  += tSTAT[1].min;
  tTOT.max  += tSTAT[1].max;
  tTOT.ave  += tSTAT[1].ave;
  tTOT.last += tSTAT[1].last;

printf("FactorLastBlock() \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_FLB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[2].min) tSTAT[2].min = node_tim.busy;
    if (node_tim.busy > tSTAT[2].max) tSTAT[2].max = node_tim.busy;
    tSTAT[2].ave += node_tim.busy;
   } /* end for i */
  tSTAT[2].last = node_tim.busy;
  tSTAT[2].ave /= (double) nprocs;
  tFAC.min  += tSTAT[2].min;
  tFAC.max  += tSTAT[2].max;
  tFAC.ave  += tSTAT[2].ave;
  tFAC.last += tSTAT[2].last;
  tTOT.min  += tSTAT[2].min;
  tTOT.max  += tSTAT[2].max;
  tTOT.ave  += tSTAT[2].ave;
  tTOT.last += tSTAT[2].last;

printf("ForwardReduce - independent blocks \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_FRIB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[3].min) tSTAT[3].min = node_tim.busy;
    if (node_tim.busy > tSTAT[3].max) tSTAT[3].max = node_tim.busy;
    tSTAT[3].ave += node_tim.busy;
   } /* end for i */
  tSTAT[3].last = node_tim.busy;
  tSTAT[3].ave /= (double) nprocs;
  tFR.min   += tSTAT[3].min;
  tFR.max   += tSTAT[3].max;
  tFR.ave   += tSTAT[3].ave;
  tFR.last  += tSTAT[3].last;
  tTOT.min  += tSTAT[3].min;
  tTOT.max  += tSTAT[3].max;
  tTOT.ave  += tSTAT[3].ave;
  tTOT.last += tSTAT[3].last;

printf("ForwardReduce - lower border \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_FRLB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[4].min) tSTAT[4].min = node_tim.busy;
    if (node_tim.busy > tSTAT[4].max) tSTAT[4].max = node_tim.busy;
    tSTAT[4].ave += node_tim.busy;
   } /* end for i */
  tSTAT[4].last = node_tim.busy;
  tSTAT[4].ave /= (double) nprocs;
  tFR.min   += tSTAT[4].min;
  tFR.max   += tSTAT[4].max;
  tFR.ave   += tSTAT[4].ave;
  tFR.last  += tSTAT[4].last;
  tTOT.min  += tSTAT[4].min;
  tTOT.max  += tSTAT[4].max;
  tTOT.ave  += tSTAT[4].ave;
  tTOT.last += tSTAT[4].last;

printf("ForwardReduce - last block \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_FR,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[5].min) tSTAT[5].min = node_tim.busy;
    if (node_tim.busy > tSTAT[5].max) tSTAT[5].max = node_tim.busy;
    tSTAT[5].ave += node_tim.busy;
   } /* end for i */
  tSTAT[5].last = node_tim.busy;
  tSTAT[5].ave /= (double) nprocs;
  tFR.min   += tSTAT[5].min;
  tFR.max   += tSTAT[5].max;
  tFR.ave   += tSTAT[5].ave;
  tFR.last  += tSTAT[5].last;
  tTOT.min  += tSTAT[5].min;
  tTOT.max  += tSTAT[5].max;
  tTOT.ave  += tSTAT[5].ave;
  tTOT.last += tSTAT[5].last;

printf("BackSubLastBlock() - last block \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_BSLB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[6].min) tSTAT[6].min = node_tim.busy;
    if (node_tim.busy > tSTAT[6].max) tSTAT[6].max = node_tim.busy;
    tSTAT[6].ave += node_tim.busy;
   } /* end for i */
  tSTAT[6].last = node_tim.busy;
  tSTAT[6].ave /= (double) nprocs;
  tBS.min   += tSTAT[6].min;
  tBS.max   += tSTAT[6].max;
  tBS.ave   += tSTAT[6].ave;
  tBS.last  += tSTAT[6].last;
  tTOT.min  += tSTAT[6].min;
  tTOT.max  += tSTAT[6].max;
  tTOT.ave  += tSTAT[6].ave;
  tTOT.last += tSTAT[6].last;

printf("BackwardSubstitution() - independent blocks \n");
fflush(stdout);
  CMMD_sync_host_with_nodes();
  for (i=0; i<nprocs; i++)
   {
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_BSIB,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    if (node_tim.busy < tSTAT[7].min) tSTAT[7].min = node_tim.busy;
    if (node_tim.busy > tSTAT[7].max) tSTAT[7].max = node_tim.busy;
    tSTAT[7].ave += node_tim.busy;
   } /* end for i */  
  tSTAT[7].last = node_tim.busy;
  tSTAT[7].ave /= (double) nprocs;
  tBS.min   += tSTAT[7].min;
  tBS.max   += tSTAT[7].max;
  tBS.ave   += tSTAT[7].ave;
  tBS.last  += tSTAT[7].last;
  tTOT.min  += tSTAT[7].min;
  tTOT.max  += tSTAT[7].max;
  tTOT.ave  += tSTAT[7].ave;
  tTOT.last += tSTAT[7].last;

  printf("\nNumber of blocks = |%d| \n",(nblocks+1));
  printf("Total number of nonzeros = |%d| \n",
           (n+nblockrow+nborderrow+(nlb*(nlb-1)/2)));
  printf("Fraction of nonzeros in the matrix = |%lf| \n\n",
          ((double)(n+nblockrow+nborderrow+(nlb*(nlb-1)/2)) / (double)((n+nlb)*(n+nlb))) );
  fflush(stdout);

  printf("\nTIMING STATISTICS \n\n");
  printf("\nFactor \t\t  tFAC.min=%lf \t tFAC.ave=%lf \t tFAC.max=%lf \t tFAC.last=%lf\n",
                tFAC.min,tFAC.ave,tFAC.max,tFAC.last);
  printf("For Red \t  tFR.min=%lf \t tFR.ave=%lf \t tFR.max=%lf \t tFR.last=%lf\n",
                tFR.min,tFR.ave,tFR.max,tFR.last);
  printf("Back Sub \t  tBS.min=%lf \t tBS.ave=%lf \t tBS.max=%lf \t tBS.last=%lf\n",
                tBS.min,tBS.ave,tBS.max,tBS.last);
  printf("Total \t\t  tTOT.min=%lf \t tTOT.ave=%lf \t tTOT.max=%lf \t tTOT.last=%lf\n\n",
                tTOT.min,tTOT.ave,tTOT.max,tTOT.last);
  for (i=0;i<8;i++)
    {
     printf("\t tSTAT[%d].min=%lf \t tSTAT[%d].ave=%lf \t tSTAT[%d].max=%lf \t tSTAT[%d].last=%lf\n",
               i,tSTAT[i].min,i,tSTAT[i].ave,i,tSTAT[i].max,i,tSTAT[i].last);
    } /* end for i */

} /* end main() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void InputData(fnameA,fnameb)
  char *fnameA,*fnameb;
{
  int c,i,j,k,l,m,iproc,jb,bufsize,ptr;
  int lastblockstart,lastblockend;
  double aii,aij;
  int endLBix,endLBkx;
  int LBPtr,EndLBPtr;
  int UBPtr,EndUBPtr;
  int Histo[33];

  char *EOFCheck;
  FILE *fd, *fopen();

  EOFCheck = (char *) calloc(80,sizeof(char));
  if (EOFCheck == NULL)
   {
    printf("can't allocate memory for EOFCheck \n");
    exit(1);
   } /* end if */

  fd = fopen(fnameA,"r");
  if (fd == NULL)
   {
    printf("can't open file %s\n",fnameA);
    exit(1);
   } /* end if */

/* (R-0) read in the processor identifier array of structures */
  fscanf(fd,"%d %d %d %d",&nprocs,&ndiag,&nlbdiag,&nb);
printf("nprocs=|%d| - ndiag=|%d| - nlbdiag=|%d| - nb=|%d|\n",nprocs,ndiag,nlbdiag,nb);
fflush(stdout);
  proc = (Proc *) calloc((nprocs+1),sizeof(Proc));
  if (proc == NULL)
   {
    printf("can't allocate memory for proc[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nprocs; i++)
   {
    fscanf(fd,"%d %d %d %d %d %d %d %d %d %d",&c,&c,
           &(proc[i].BlockPtr),
           &(proc[i].DiagPtr),
           &(proc[i].BlockRowPtr),
           &(proc[i].LowerBorderPtr),
           &(proc[i].BorderRowPtr),
           &(proc[i].BlockColPtr),
           &(proc[i].UpperBorderPtr),
           &(proc[i].BorderColPtr));
   } /* end for i */


/* (R-1) read in the block identifier array of structures */
  fscanf(fd,"%d",&nblocks);
printf("nblocks=|%d| \n",nblocks);
fflush(stdout);
  blocks = (Block *) calloc((nblocks+1),sizeof(Block));
  if (blocks == NULL)
   {
    printf("can't allocate memory for blocks[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nblocks; i++)
   {
    fscanf(fd,"%d %d %d %d %d %d %d",&c,
            &(blocks[i].BlockID),
            &(blocks[i].BlockSz),
            &(blocks[i].ProcNum),
            &(blocks[i].DiagPtr),
            &(blocks[i].LBPtr),
            &(blocks[i].UBPtr) );
   } /* end for i */

/* (R-2) read in the matrix diagonal array of structures */
  fscanf(fd,"%d",&n);
printf("n=|%d| \n",n);
fflush(stdout);
  diag = (Diag *) calloc((n+1),sizeof(Diag));
  if (diag == NULL)
   {
    printf("can't allocate memory for diag[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=n; i++)
   {
    fscanf(fd,"%d %lf %d %d %d %d %d %d",&c,
            &(diag[i].Value),
            &(diag[i].PtrLix),
            &(diag[i].NextLxk),
            &(diag[i].NextLBxk),
            &(diag[i].PtrUxj),
            &(diag[i].NextUkx),
            &(diag[i].NextUBkx) );
   } /* end for i */

/* (R-3) read in the block row storage array of structures */
  fscanf(fd,"%d",&nblockrow);
printf("nblockrow=|%d| \n",nblockrow);
fflush(stdout);
  blockrow = (BlockRow *) calloc((nblockrow+1),sizeof(BlockRow));
  if (blockrow == NULL)
   {
    printf("can't allocate memory for blockrow[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nblockrow; i++)
   {
    fscanf(fd,"%d %lf %d %d %d",&c,
            &(blockrow[i].Value),
            &(blockrow[i].i),
            &(blockrow[i].j),
            &(blockrow[i].NextLxk) );
   } /* end for i */

/* (R-4) read in the block column storage array of structures */
  fscanf(fd,"%d",&nblockcol);
printf("nblockcol=|%d| \n",nblockcol);
fflush(stdout);
  blockcol = (BlockCol *) calloc((nblockcol+1),sizeof(BlockCol));
  if (blockcol == NULL)
   {
    printf("can't allocate memory for blockcol[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nblockcol; i++)
   {
    fscanf(fd,"%d %lf %d %d %d",&c,
            &(blockcol[i].Value),
            &(blockcol[i].i),
            &(blockcol[i].j),
            &(blockcol[i].NextUkx) );
   } /* end for i */

/* (R-5) read in the lower border row storage array of structures */
  fscanf(fd,"%d",&nlborder);
printf("nlborder=|%d| \n",nlborder);
fflush(stdout);
  lborder = (LowerBorder *) calloc((nlborder+1),sizeof(LowerBorder));
  if (lborder == NULL)
   {
    printf("can't allocate memory for lborder[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nlborder; i++)
   {
    fscanf(fd,"%d %d %d",&c,
            &(lborder[i].i),
            &(lborder[i].PtrLBix) );
   } /* end for i */

/* (R-6) read in the upper border column storage array of structures */
  fscanf(fd,"%d",&nuborder);
printf("nuborder=|%d| \n",nuborder);
fflush(stdout);
  uborder = (UpperBorder *) calloc((nuborder+1),sizeof(UpperBorder));
  if (uborder == NULL)
   {
    printf("can't allocate memory for uborder[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nuborder; i++)
   {
    fscanf(fd,"%d %d %d",&c,
            &(uborder[i].j),
            &(uborder[i].PtrUBxj) );
   } /* end for i */

/* (R-7) read in the border row storage array of structures */
  fscanf(fd,"%d",&nborderrow);
printf("nborderrow=|%d| \n",nborderrow);
fflush(stdout);
  borderrow = (BorderRow *) calloc((nborderrow+1),sizeof(BorderRow));
  if (borderrow == NULL)
   {
    printf("can't allocate memory for borderrow[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nborderrow; i++)
   {
    fscanf(fd,"%d %lf %d %d %d",&c,
            &(borderrow[i].Value),
            &(borderrow[i].i),
            &(borderrow[i].j),
            &(borderrow[i].NextLBxk) );
   } /* end for i */

/* (R-8) read in the border column storage array of structures */
  fscanf(fd,"%d",&nbordercol);
printf("nbordercol=|%d| \n",nbordercol);
fflush(stdout);
  bordercol = (BorderCol *) calloc((nbordercol+1),sizeof(BorderCol));
  if (bordercol == NULL)
   {
    printf("can't allocate memory for bordercol[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nbordercol; i++)
   {
    fscanf(fd,"%d %lf %d %d %d",&c,
            &(bordercol[i].Value),
            &(bordercol[i].i),
            &(bordercol[i].j),
            &(bordercol[i].NextUBkx) );
   } /* end for i */

/* (I-0) load the block structure buffers for the independent blocks */
  bbuf.nprocs     = nprocs;
  bbuf.n          = n;
  bbuf.nblocks    = nblocks;
  bbuf.nblockrow  = nblockrow;
  bbuf.nblockcol  = nblockcol;
  bbuf.nlborder   = nlborder;
  bbuf.nuborder   = nuborder;
  bbuf.nborderrow = nborderrow;
  bbuf.nbordercol = nbordercol;
  bbuf.nlb        = nlbdiag;
  bbuf.nb         = nb;
  CMMD_bc_from_host(&bbuf,sizeof(BBuf));

/* (I-1) load the block structure buffers for the independent blocks */
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].BlockPtr;
    m = proc[i+1].BlockPtr;
    npblocks = (m-l);
printf("load the block structure buffers for the independent blocks npblocks=%d\n",
              npblocks);
fflush(stdout);
    CMMD_send(i,SEND_NPBLOCKS,&npblocks,sizeof(int));
    pblocks = (Block *) calloc((npblocks+1),sizeof(Block));
    if (pblocks == NULL)
     {
      printf("can't allocate memory for pblocks[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pblocks[0].BlockID = 0;
    pblocks[0].BlockSz = 0;
    pblocks[0].DiagPtr = blocks[l].DiagPtr;
    pblocks[0].LBPtr   = 0;
    pblocks[0].UBPtr   = 0;

printf("\ti=%d - l=%d - m=%d - pblocks[0].DiagPtr=%d \n",i,l,m,pblocks[0].DiagPtr);
fflush(stdout);

    for (j=1; j<=npblocks; j++)
     {
      pblocks[j].BlockID = blocks[j+l].BlockID;
      pblocks[j].BlockSz = blocks[j+l].BlockSz;
      pblocks[j].DiagPtr = blocks[j+l].DiagPtr;
      if (blocks[j+l].LBPtr == 0)
        pblocks[j].LBPtr = 0;
      else
        pblocks[j].LBPtr = blocks[j+l].LBPtr - blocks[l].LBPtr;
      if (blocks[j+l].UBPtr == 0)
        pblocks[j].UBPtr = 0;
      else
        pblocks[j].UBPtr = blocks[j+l].UBPtr - blocks[l].UBPtr;
     } /* end for j */
/*
printf("send the block structure buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PBLOCKS,pblocks,((npblocks+1)*sizeof(Block)));
    free(pblocks); 
   } /* end for i */

/* (I-2) load the matrix diagonal structure buffers for the independent blocks */
printf("load the matrix diagonal structure buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].DiagPtr;
    m = proc[i+1].DiagPtr;
    np = (m-l);
    CMMD_send(i,SEND_NP,&np,sizeof(int));
    pdiag = (Diag *) calloc((np+1),sizeof(Diag));
    if (pdiag == NULL)
     {
      printf("can't allocate memory for pdiag[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pdiag[0].Value    = 0.0;
    pdiag[0].PtrLix   = 0;
    pdiag[0].NextLxk  = 0;
    pdiag[0].NextLBxk = 0;
    pdiag[0].PtrUxj   = 0;
    pdiag[0].NextUkx  = 0;
    pdiag[0].NextUBkx = 0;

    for (j=1; j<=np; j++)
     {
      pdiag[j].Value    = diag[j+l].Value;

      pdiag[j].PtrLix   = diag[j+l].PtrLix - proc[i].BlockRowPtr;
      if (diag[j+l].NextLxk == 0)
        pdiag[j].NextLxk = 0;
      else
        pdiag[j].NextLxk  = diag[j+l].NextLxk - proc[i].BlockRowPtr;
      if (diag[j+l].NextLBxk == 0)
        pdiag[j].NextLBxk = 0;
      else
        pdiag[j].NextLBxk = diag[j+l].NextLBxk - proc[i].BorderRowPtr;

      pdiag[j].PtrUxj   = diag[j+l].PtrUxj - proc[i].BlockColPtr;
      if (diag[j+l].NextUkx == 0)
        pdiag[j].NextUkx = 0;
      else
        pdiag[j].NextUkx  = diag[j+l].NextUkx - proc[i].BlockColPtr;
      if (diag[j+l].NextUBkx == 0)
        pdiag[j].NextUBkx = 0;
      else
        pdiag[j].NextUBkx = diag[j+l].NextUBkx - proc[i].BorderColPtr;
     } /* end for j */
/*
printf("send the matrix diagonal structure buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PDIAG,pdiag,((np+1)*sizeof(Diag)));
    free(pdiag);
   } /* end for i */

/* (I-3) load the block row storage structure buffers for the independent blocks */
printf("load the block row storage structure buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].BlockRowPtr;
    m = proc[i+1].BlockRowPtr;
    npblockrow = (m-l);
    CMMD_send(i,SEND_NPBLOCKROW,&npblockrow,sizeof(int));
    pblockrow = (BlockRow *) calloc((npblockrow+1),sizeof(BlockRow));
    if (pblockrow == NULL)
     {
      printf("can't allocate memory for pblockrow[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pblockrow[0].Value   = 0.0;
    pblockrow[0].i       = 0;
    pblockrow[0].j       = 0;
    pblockrow[0].NextLxk = 0;
    for (j=1; j<=npblockrow; j++)
     {
      pblockrow[j].Value   = blockrow[j+l].Value;
      pblockrow[j].i       = blockrow[j+l].i;
      pblockrow[j].j       = blockrow[j+l].j;
      if (blockrow[j+l].NextLxk == 0)
        pblockrow[j].NextLxk = 0;
      else
        pblockrow[j].NextLxk = blockrow[j+l].NextLxk - l;
     } /* end for j */
/*
printf("send the block row storage structure buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PBLOCKROW,pblockrow,((npblockrow+1)*sizeof(BlockRow)));
    free(pblockrow);
   } /* end for i */

/* (I-4) load the block column storage structure buffers for the independent blocks */
printf("load the block column storage structure buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].BlockColPtr;
    m = proc[i+1].BlockColPtr;
    npblockcol = (m-l);
    CMMD_send(i,SEND_NPBLOCKCOL,&npblockcol,sizeof(int));
    pblockcol = (BlockCol *) calloc((npblockcol+1),sizeof(BlockCol));
    if (pblockcol == NULL)
     {
      printf("can't allocate memory for pblockcol[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pblockcol[0].Value   = 0.0;
    pblockcol[0].i       = 0;
    pblockcol[0].j       = 0;
    pblockcol[0].NextUkx = 0;
    for (j=1; j<=npblockcol; j++)
     {
      pblockcol[j].Value   = blockcol[j+l].Value;
      pblockcol[j].i       = blockcol[j+l].i;
      pblockcol[j].j       = blockcol[j+l].j;
      if (blockcol[j+l].NextUkx == 0)
        pblockcol[j].NextUkx = 0;
      else
        pblockcol[j].NextUkx = blockcol[j+l].NextUkx - l;
     } /* end for j */
/*
printf("send the block column storage structure buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PBLOCKCOL,pblockcol,((npblockcol+1)*sizeof(BlockCol)));
    free(pblockcol);
   } /* end for i */

/* (I-5) load the lower border row storage buffers for the independent blocks */
printf("load the lower border row storage buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].LowerBorderPtr;
    m = proc[i+1].LowerBorderPtr;
    nplborder = (m-l);
    CMMD_send(i,SEND_NPLBORDER,&nplborder,sizeof(int));
    plborder = (LowerBorder *) calloc((nplborder+1),sizeof(LowerBorder));

printf("iproc=%d - l=%d - m=%d - nplborder=%d \n",i,l,m,nplborder);
fflush(stdout);

    if (plborder == NULL)
     {
      printf("can't allocate memory for plborder[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    plborder[0].i       = 0;
    plborder[0].PtrLBix = 0;
    for (j=1; j<=nplborder; j++)
     {
      plborder[j].i       = lborder[j+l].i;
      plborder[j].PtrLBix = lborder[j+l].PtrLBix - proc[i].BorderRowPtr;
     } /* end for j */
/*
printf("send the lower border row storage buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PLBORDER,plborder,((nplborder+1)*sizeof(LowerBorder)));
    free(plborder);
   } /* end for i */

/* (I-6) load the upper border column storage buffers for the independent blocks */
printf("load the upper border column storage buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].UpperBorderPtr;
    m = proc[i+1].UpperBorderPtr;
    npuborder = (m-l);
    CMMD_send(i,SEND_NPUBORDER,&npuborder,sizeof(int));
    puborder = (UpperBorder *) calloc((npuborder+1),sizeof(UpperBorder));

printf("iproc=%d - l=%d - m=%d - npuborder=%d \n",i,l,m,npuborder);
fflush(stdout);

    if (puborder == NULL)
     {
      printf("can't allocate memory for puborder[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    puborder[0].j       = 0;
    puborder[0].PtrUBxj = 0;
    for (j=1; j<=npuborder; j++)
     {
      puborder[j].j       = uborder[j+l].j;
      puborder[j].PtrUBxj = uborder[j+l].PtrUBxj - proc[i].BorderColPtr;
     } /* end for j */
/*
printf("send the upper border row storage buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PUBORDER,puborder,((npuborder+1)*sizeof(UpperBorder)));
    free(puborder);
   } /* end for i */

/* (I-7) load the border row structure buffers for the independent blocks */
printf("load the border row structure buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].BorderRowPtr;
    m = proc[i+1].BorderRowPtr;
    npborderrow = (m-l);
    CMMD_send(i,SEND_NPBORDERROW,&npborderrow,sizeof(int));
    pborderrow = (BorderRow *) calloc((npborderrow+1),sizeof(BorderRow));
    if (pborderrow == NULL)
     {
      printf("can't allocate memory for pborderrow[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pborderrow[0].Value    = 0.0;
    pborderrow[0].i        = 0;
    pborderrow[0].j        = 0;
    pborderrow[0].NextLBxk = 0;
    for (j=1; j<=npborderrow; j++)
     {
      pborderrow[j].Value    = borderrow[j+l].Value;
      pborderrow[j].i        = borderrow[j+l].i;
      pborderrow[j].j        = borderrow[j+l].j;
      if (borderrow[j+l].NextLBxk == 0)
       pborderrow[j].NextLBxk = 0;
      else
       pborderrow[j].NextLBxk = borderrow[j+l].NextLBxk - l;
     } /* end for j */
/*
printf("send the border row structure buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PBORDERROW,pborderrow,((npborderrow+1)*sizeof(BorderRow)));
    free(pborderrow);
   } /* end for i */

/* (I-8) load the border column structure buffers for the independent blocks */
printf("load the border column structure buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].BorderColPtr;
    m = proc[i+1].BorderColPtr;
    npbordercol = (m-l);
    CMMD_send(i,SEND_NPBORDERCOL,&npbordercol,sizeof(int));
    pbordercol = (BorderCol *) calloc((npbordercol+1),sizeof(BorderCol));
    if (pbordercol == NULL)
     {
      printf("can't allocate memory for pbordercol[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pbordercol[0].Value    = 0.0;
    pbordercol[0].i        = 0;
    pbordercol[0].j        = 0;
    pbordercol[0].NextUBkx = 0;
    for (j=1; j<=npbordercol; j++)
     {
      pbordercol[j].Value    = bordercol[j+l].Value;
      pbordercol[j].i        = bordercol[j+l].i;
      pbordercol[j].j        = bordercol[j+l].j;
      if (bordercol[j+l].NextUBkx == 0)
       pbordercol[j].NextUBkx = 0;
      else
       pbordercol[j].NextUBkx = bordercol[j+l].NextUBkx - l;
     } /* end for j */
/*
printf("send the border column structure buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PBORDERCOL,pbordercol,((npbordercol+1)*sizeof(BorderCol)));
    free(pbordercol);
   } /* end for i */


/* read and send the data in the last diagonal block - data will be prepared in partition.c */

/* (II-1) read and send the last block diagonal structure */
  for (iproc=0;iproc<nprocs;iproc++)
    {
  /* read */
     fscanf(fd,"%d %d",&nplbdiag,&nb);
printf("\tiproc=|%d| nplbdiag=|%d| \n",iproc,nplbdiag);
fflush(stdout);
     plbdiag = (LBDiag *) calloc((nplbdiag+1),sizeof(LBDiag));
     if (plbdiag == NULL)
      {
       printf("can't allocate memory for plbdiag[] \n");
       exit(1);
      } /* end if */
     for (i=1; i<=nplbdiag; i++)
      {
       fscanf(fd,"%d %lf %d %d %d %d %d",&c,
               &(plbdiag[i].Value),
               &(plbdiag[i].PtrUix),
               &(plbdiag[i].PtrUxj),
               &(plbdiag[i].PtrLix),
               &(plbdiag[i].PtrLxj),
               &(plbdiag[i].k) );
      } /* end for i */
     plbdiag[0].Value  = 0.0;
     plbdiag[0].PtrUix = 0;
     plbdiag[0].PtrUxj = 0;
     plbdiag[0].PtrLix = 0;
     plbdiag[0].PtrLxj = 0;
     plbdiag[0].k      = 0;
  /* send */
     CMMD_send(iproc,SEND_NLBDIAG,&nplbdiag,sizeof(int));
     CMMD_send(iproc,SEND_NB,&nb,sizeof(int));
     CMMD_send(iproc,SEND_LBDIAG,plbdiag,((nplbdiag+1)*sizeof(LBDiag)));
     free(plbdiag);

/* (II-2) read and send the last block rows in the lower diagonal structure */
  /* read */
     fscanf(fd,"%d",&nplbrowL);
printf("\tiproc=|%d| nplbrowL=|%d| \n",iproc,nplbrowL);
fflush(stdout);
     plbrowL = (LBRow *) calloc((nplbrowL+1),sizeof(LBRow));
     if (plbrowL == NULL)
      {
       printf("can't allocate memory for plbrowL[] \n");
       exit(1);
      } /* end if */
     for (i=1; i<=nplbrowL; i++)
      {
       fscanf(fd,"%d %lf %d %d",&c,
               &(plbrowL[i].Value),
               &(plbrowL[i].i),
               &(plbrowL[i].j) );
      } /* end for i */
     plbrowL[0].Value  = 0.0;
     plbrowL[0].i      = 0;
     plbrowL[0].j      = 0;
  /* send */
     CMMD_send(iproc,SEND_NLBROWL,&nplbrowL,sizeof(int));
     CMMD_send(iproc,SEND_LBROWL,plbrowL,((nplbrowL+1)*sizeof(LBRow)));
     free(plbrowL);

/* (II-3) read and send the last block rows in the upper diagonal structure */
  /* read */
     fscanf(fd,"%d",&nplbrowU);
printf("\tiproc=|%d| nplbrowU=|%d| \n",iproc,nplbrowU);
fflush(stdout);
     plbrowU = (LBRow *) calloc((nplbrowU+1),sizeof(LBRow));
     if (plbrowU == NULL)
      {
       printf("can't allocate memory for plbrowU[] \n");
       exit(1);
      } /* end if */
     for (i=1; i<=nplbrowU; i++)
      {
       fscanf(fd,"%d %lf %d %d",&c,
               &(plbrowU[i].Value),
               &(plbrowU[i].i),
               &(plbrowU[i].j) );
      } /* end for i */
     plbrowU[0].Value  = 0.0;
     plbrowU[0].i      = 0;
     plbrowU[0].j      = 0;
  /* send */
     CMMD_send(iproc,SEND_NLBROWU,&nplbrowU,sizeof(int));
     CMMD_send(iproc,SEND_LBROWU,plbrowU,((nplbrowU+1)*sizeof(LBRow)));
     free(plbrowU);

/* (II-4) read and send the last block lower diagonal column pointers structure  */
  /* read */
     fscanf(fd,"%d",&nplbcolL);
printf("\tiproc=|%d| nplbcolL=|%d| \n",iproc,nplbcolL);
fflush(stdout);
     plbcolL = (LBCol *) calloc((nplbcolL+1),sizeof(LBCol));
     if (plbcolL == NULL)
      {
       printf("can't allocate memory for plbcolL[] \n");
       exit(1);
      } /* end if */
     for (i=1; i<=nplbcolL; i++)
      {
       fscanf(fd,"%d %d",&c,
               &(plbcolL[i].PtrNext) );
      } /* end for i */
     plbcolL[0].PtrNext = 0;
  /* send */
     CMMD_send(iproc,SEND_LBCOLL,plbcolL,((nplbcolL+1)*sizeof(LBCol)));
     free(plbcolL);

/* (II-5) read and send the last block upper diagonal column pointers structure */
  /* read */
     fscanf(fd,"%d",&nplbcolU);
printf("\tiproc=|%d| nplbcolU=|%d| \n",iproc,nplbcolU);
fflush(stdout);
     plbcolU = (LBCol *) calloc((nplbcolU+1),sizeof(LBCol));
     if (plbcolU == NULL)
      {
       printf("can't allocate memory for plbcolU[] \n");
       exit(1);
      } /* end if */
     for (i=1; i<=nplbcolU; i++)
      {
       fscanf(fd,"%d %d",&c,
               &(plbcolU[i].PtrNext) );
      } /* end for i */
     plbcolU[0].PtrNext = 0;
  /* send */
     CMMD_send(iproc,SEND_LBCOLU,plbcolU,((nplbcolU+1)*sizeof(LBCol)));
     free(plbcolU);

/* (II-6) read and send the last block all columns column pointers structure */
  /* read */
     fscanf(fd,"%d",&nplbdiagcolpt);
printf("\tiproc=|%d| nplbdiagcolpt=|%d| \n",iproc,nplbdiagcolpt);
fflush(stdout);
     plbdiagcolpt = (LBDiagColPt *) calloc((nplbdiagcolpt+1),sizeof(LBDiagColPt));
     if (plbdiagcolpt == NULL)
      {
       printf("can't allocate memory for plbdiagcolpt[] \n");
       exit(1);
      } /* end if */
     for (i=1; i<=nplbdiagcolpt; i++)
      {
       fscanf(fd,"%d %d %d %d",&c,
               &(plbdiagcolpt[i].k),
               &(plbdiagcolpt[i].PtrUxj),
               &(plbdiagcolpt[i].PtrLxj) );
      } /* end for i */
     plbdiagcolpt[0].k      = 0;
     plbdiagcolpt[0].PtrUxj = 0;
     plbdiagcolpt[0].PtrLxj = 0;
  /* send */
     CMMD_send(iproc,SEND_LBDIAGCOLPT,plbdiagcolpt,((nplbdiagcolpt+1)*sizeof(LBDiagColPt)));
     free(plbdiagcolpt);
   } /* end for iproc */

/* (II-7) read and send the buffer sizes when receiving the broadcasts of data when factoring the last block */
  /* read */
     fscanf(fd,"%d",&nlbbufsizeFac);
printf("\tiproc=|%d| nlbbufsizeFac=|%d| \n",iproc,nlbbufsizeFac);
fflush(stdout);
     lbbufsizeFac = (int *) calloc((nlbbufsizeFac+1),sizeof(int));
     if (lbbufsizeFac == NULL)
      {
       printf("can't allocate memory for lbbufsizeFac[] \n");
       exit(1);
      } /* end if */
     for (i=1; i<=nlbbufsizeFac; i++)
      {
       fscanf(fd,"%d %d",&c,
               &(lbbufsizeFac[i]) );
      } /* end for i */
     lbbufsizeFac[0] = 0.0;
  /* send */
  for (iproc=0;iproc<nprocs;iproc++)
    {
      CMMD_send(iproc,SEND_NLBBUFSIZEFAC,&nlbbufsizeFac,sizeof(int));
      CMMD_send(iproc,SEND_LBBUFSIZEFAC,lbbufsizeFac,((nlbbufsizeFac+1)*sizeof(int)));
      free(lbbufsizeFac);
    } /* end for iproc */

/* read in a character string to identify EOF */
  fscanf(fd,"%s",EOFCheck);
  printf("Input_A_matrix - |%s| \n",EOFCheck);
  fflush(stdout);

  fclose(fd);

  fd = fopen(fnameb,"r");
  if (fd == NULL)
   {
    printf("can't open file %s\n",fnameb);
    exit(1);
   } /* end if */

/* (III-1) read in the b vector */
printf("read in the b vector \n");
fflush(stdout);
  b = (double *) calloc((n+1),sizeof(double));
  if (b == NULL)
   {
    printf("can't allocate memory for b[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=n; i++)
   {
    fscanf(fd,"%d %lf",&c,&(b[i]));
   } /* end for i */
/* load the b vector buffers for the independent blocks */
printf("load the b vector buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].DiagPtr;
    m = proc[i+1].DiagPtr;
    np = (m-l);

printf("\tl=%d - m=%d - np=%d \n",l,m,np);
fflush(stdout);

    pb = (double *) calloc((np+1),sizeof(double));
    if (pb == NULL)
     {
      printf("can't allocate memory for pb[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    for (j=1; j<=np; j++)
     {
      pb[j] = b[j+l];
     } /* end for j */
/*
printf("load the b vector buffers for the independent blocks \n");
fflush(stdout);
*/
    CMMD_send(i,SEND_PB,pb,((np+1)*sizeof(double)));
    free(pb);
   } /* end for i */

/* (III-2) read in the lb vector */
  nlb = nlbdiag;
  lb = (double *) calloc((nlb+1),sizeof(double));
  if (lb == NULL)
   {
    printf("can't allocate memory for lb[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nlb; i++)
   {
    fscanf(fd,"%d %lf",&c,&(lb[i]));
   } /* end for i */

/* read in a character string to identify EOF */
  fscanf(fd,"%s",EOFCheck);
  printf("Input_b_vector - |%s| \n",EOFCheck);
  fflush(stdout);

  fclose(fd);

/* load the lb vector onto the processors */
printf("load the lb vector onto the processors - nlb=%d \n",nlb);
fflush(stdout);
  iproc = 0;
  for (j=1; j<=nlb; j += nb)
   {
    jb = min((nlb-j+1),nb);
    bufsize = (jb*sizeof(double));
/*
printf("send the lb vector onto the processors iproc=%d - j=%d - jb=%d \n",iproc,j,jb);
fflush(stdout);
*/
    CMMD_send(iproc,SEND_LB,&(lb[j]),bufsize);
/*
printf("send the lb vector onto the processors - after send - j=%d \n",j);
fflush(stdout);
*/
    iproc = (iproc+1)%nprocs;
   } /* end for j */

/* (IV-1) send the row/col locations of the first and last locations in the
   last block */
printf("send the row/col locations of lastblockstart \n");
fflush(stdout);
  lastblockstart = blocks[nblocks].DiagPtr + 1;
  for (iproc=0;iproc<nprocs;iproc++)
    CMMD_send(iproc,SEND_LASTBKST,&lastblockstart,sizeof(int));
/*
printf("send the row/col locations of lastblockend \n");
fflush(stdout);
*/
  lastblockend   = blocks[nblocks].DiagPtr + nlb;
  for (iproc=0;iproc<nprocs;iproc++)
    CMMD_send(iproc,SEND_LASTBKND,&lastblockend,sizeof(int));
printf("lastblockend=%d - lastblockstart=%d \n",lastblockend,lastblockstart);
fflush(stdout);


/* (IV-2) calculate and send the number of updates for factoring the last block */
  nupdatesULB = (int *) calloc((nprocs+1),sizeof(int));
  if (nupdatesULB == NULL)
   {
    printf("can't allocate memory for nupdatesULB[] \n");
    exit(1);
   } /* end if */
  for (i=0;i<=nprocs;i++) nupdatesULB[i] = 0;

  for (l=1; l<=nblocks; l++)
   {
/*
printf("\t\tl=%d - nblocks=%d \n",l,nblocks);
fflush(stdout);
*/
    LBPtr = (blocks[l-1].LBPtr + 1);
    EndLBPtr = blocks[l].LBPtr;

/* loop over all non-zero lower border rows in this block ---
   also visit all non-zero upper border columns in the corresponding
   block location --- visit them in a pair of nested while loops */
/* outer loop */
    while (LBPtr <= EndLBPtr)
      {
       i = lborder[LBPtr].i - lastblockstart;
       UBPtr = (blocks[l-1].UBPtr + 1);
       EndUBPtr = blocks[l].UBPtr;
/* inner loop */
       while (UBPtr <= EndUBPtr)
         {
          j = uborder[UBPtr].j - lastblockstart;
          iproc = ((i%(nprocs*nb))/nb);
          (nupdatesULB[iproc+1])++;
/*
if (iproc == 10) {
printf("i=%d - j=%d - iproc=%d - nupdatesULB[%d]=%d - LBPtr=%d - UBPtr=%d \n",
        i,j,iproc,(iproc+1),nupdatesULB[iproc+1],LBPtr,UBPtr);
fflush(stdout); }
*/
          UBPtr++;
         } /* end inner loop */
       LBPtr++;
      } /* end outer loop */
    } /* end for l */

  for (i=0;i<=nprocs;i++)
    printf("\t\tiproc=%d - nupdatesULB[%d]=%d \n",i,i,nupdatesULB[i]);
  fflush(stdout);

  bufsize = ((nprocs+1)*sizeof(int));
  for (iproc=0;iproc<nprocs;iproc++)
    {
     CMMD_send(iproc,SEND_NUPDATEULB,nupdatesULB,bufsize);
    } /* end for iproc */

/* (IV-3) calculate and send the number of updates for forward reducing the last block */
  nupdatesFRB = (int *) calloc((nprocs+1),sizeof(int));
  if (nupdatesFRB == NULL)
   {
    printf("can't allocate memory for nupdatesFRB[] \n");
    exit(1);
   } /* end if */
  for (i=0;i<=nprocs;i++) nupdatesFRB[i] = 0;

  for (l=1; l<=nblocks; l++)
   {
/*
printf("\t\tl=%d - nblocks=%d \n",l,nblocks);
fflush(stdout);
*/
    LBPtr = (blocks[l-1].LBPtr + 1);
    EndLBPtr = blocks[l].LBPtr;

/* loop over all non-zero lower border rows in this block */
    while (LBPtr <= EndLBPtr) 
      {
/*
printf("\t\t\t\t\tLBPtr=%d - EndLBPtr=%d \n",LBPtr,EndLBPtr);
fflush(stdout);
*/
       i = lborder[LBPtr].i - lastblockstart + 1;
       iproc = (((i-1)%(nprocs*nb))/nb);
       nupdatesFRB[iproc+1]++;
       LBPtr++;
      } /* end outer loop */
    } /* end for l */

  for (i=0;i<=nprocs;i++)
    printf("\t\tiproc=%d - nupdatesFRB[%d]=%d \n",i,i,nupdatesFRB[i]);
  fflush(stdout);

  bufsize = ((nprocs+1)*sizeof(int));
  for (iproc=0;iproc<nprocs;iproc++)
    {
     CMMD_send(iproc,SEND_NUPDATEFRB,nupdatesFRB,bufsize);
    } /* end for iproc */

/* (IV) allocate memory for the remaining variables */
  tSTAT = (CMMD_Timer *) calloc(11,sizeof(CMMD_Timer));
  if (tSTAT == NULL)
   {
    printf("can't allocate memory for tSTAT[] \n");
    exit(1);
   } /* end if */

  for (i=0;i<=10;i++)
    {
     tSTAT[i].min  = 100000000.0;
     tSTAT[i].ave  = 0.0;
     tSTAT[i].max  = 0.0;
     tSTAT[i].last = 0.0;
    } /* end for i */

  tFAC.min  = 0.0;
  tFAC.ave  = 0.0;
  tFAC.max  = 0.0;
  tFAC.last = 0.0;

  tFR.min  = 0.0;
  tFR.ave  = 0.0;
  tFR.max  = 0.0;
  tFR.last = 0.0;

  tBS.min  = 0.0;
  tBS.ave  = 0.0;
  tBS.max  = 0.0;
  tBS.last = 0.0;

  tTOT.min  = 0.0;
  tTOT.ave  = 0.0;
  tTOT.max  = 0.0;
  tTOT.last = 0.0;

} /* end InputData() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
int min(i,j)
 int i,j;
{
  if (i<=j) return i;
   else return j;

} /* end min() */
